import json
from typing import Dict
import os

def model_config(model_path:str) -> Dict:
    config_file = model_path+"/config.json"
    if os.path.exists(config_file):
        with open(config_file) as f:
            config_info = json.load(f)
    else:
        raise Exception("Model Config json is not available")

    return config_info





from loguru import logger
from skl2onnx import convert_sklearn
from skl2onnx.common.data_types import FloatTensorType
from joblib import load
import os
import sclblonnx as so
from .config import model_config
from .onnxexporter import ONNXExporter
from .pooler import MeanPoolingOnnx


def convert_onnx(model_path:str,output_dir:str,quantize:bool=False,opset_version=12):
    
    
    # Check whether the specified output directory exists or not
    isExist = os.path.exists(output_dir)
    if not isExist:
        # Create a new directory because it does not exist
        os.makedirs(output_dir)
    onnx_exporter = ONNXExporter()
    logger.info("Exporting to ONNX...")
    pooled_model_pytorch = MeanPoolingOnnx(model_path)
    # logger.info("Mean Pooling Model is built successfully!")
    # logger.info("Exporting Pooled Model to ONNX")
    pooled_model_onnx = onnx_exporter(model=pooled_model_pytorch,tokenizer_path=model_path,output=output_dir+"/embeddings.onnx",quantize=quantize,opset=opset_version)
    # logger.info("ONNX Model Successfully built in the directory '{}'".format(output_dir))
    # logger.info("Exporting model_head.pkl to ONNX...")
    # Get the input shape of the model
    model_info = model_config(model_path)
    output_shape = model_info.get('hidden_size',None)
    if not output_shape:
        raise Exception("Hidden Size detail not available in model config.json")
    initial_type = [('float_input', FloatTensorType([output_shape]))]
    try:
        model_head = load(model_path+'/model_head.pkl')
    except:
        logger.error("model_head.pkl file is not found in the model path directory")

    model_head_onnx = convert_sklearn(model_head, initial_types=initial_type)

    with open(output_dir+"/model_head.onnx", "wb") as f:
        f.write(model_head_onnx.SerializeToString())
    # logger.info("Model Head Exported to ONNX Successfully")

    # logger.info("Merging both Onnx Models...")
    # Read both the onnx files
    model_graph = so.graph_from_file(output_dir+"/embeddings.onnx") 
    head_graph = so.graph_from_file(output_dir+"/model_head.onnx")
    merged_model = so.merge(sg1=model_graph,sg2=head_graph,io_match=[("embeddings","float_input")],complete=False,_verbose=False)
    so.graph_to_file(merged_model,output_dir+'/setfit-model.onnx')
    # Deleting individual models
    try:
        os.remove(output_dir+"/embeddings.onnx")
        os.remove(output_dir+"/model_head.onnx")
    except:
        pass
    logger.info("ONNX Model Successfully Built in Directory:{}".format(output_dir))





from torch import nn
import torch
from transformers import AutoModel


class MeanPooling(nn.Module):
    """ MeanPooling Module for Sentence Transformer
    """
    def __init__(self,path:str):
        super().__init__()
        self.model = AutoModel.from_pretrained(path)
        

    def forward(self,**inputs):
        tokens = self.model(**inputs)[0]
        mask = inputs["attention_mask"]
        mask = mask.unsqueeze(-1).expand(tokens.size()).float()
        return torch.sum(tokens * mask, 1) / torch.clamp(mask.sum(1), min=1e-9)

class MeanPoolingOnnx(MeanPooling):
    """
    Extends MeanPooling class to name inputs to model, which is required
    to export to ONNX.
    """

    # pylint: disable=W0221
    def forward(self, input_ids=None, attention_mask=None, token_type_ids=None):
        # Build list of arguments dynamically since some models take token_type_ids
        # and others don't
        inputs = {"input_ids": input_ids, "attention_mask": attention_mask}
        if token_type_ids is not None:
            inputs["token_type_ids"] = token_type_ids

        return super().forward(**inputs)


from torch.onnx import export
from itertools import chain
from collections import OrderedDict
from torch import nn
from .pooler import MeanPoolingOnnx
from transformers import AutoTokenizer
from io import BytesIO

# Check for onnxruntime 
try:
    from onnxruntime import GraphOptimizationLevel, InferenceSession, SessionOptions
    from onnxruntime.quantization import quantize_dynamic

    ONNX_RUNTIME = True
except ImportError:
    ONNX_RUNTIME = False



class ONNXExporter:
    """
        Exports the Sentence Transformer Model to ONNX Graph.
    """

    def __call__(self, model:MeanPoolingOnnx, tokenizer_path:str,output=None, quantize=False, opset=12):
        """
        Exports the Pooled Sentence Transformer model to ONNX.

        Args:
            model: Sentence transformer nn.Module with mean pooling layer
            output: optional output model path, defaults to return byte array if None
            quantize: if model should be quantized (requires onnx to be installed), defaults to False
            opset: onnx opset, defaults to 12

        Returns:
            path to model output or model as bytes depending on output parameter
        """

        #Model Inputs
        inputs = OrderedDict(
            [
                ("input_ids", {0: "batch", 1: "sequence"}),
                ("attention_mask", {0: "batch", 1: "sequence"}),
                ("token_type_ids", {0: "batch", 1: "sequence"}),
            ]
        )
        #Model Outputs
        outputs = OrderedDict({"embeddings": {0: "batch", 1: "sequence"}})

        # Tokenizer
        tokenizer = AutoTokenizer.from_pretrained(tokenizer_path)

        # Generate dummy inputs
        dummy = dict(tokenizer(["test inputs"], return_tensors="pt"))

        # Default to BytesIO if no output file provided
        output = output if output else BytesIO()

        # Export model to ONNX
        export(
            model,
            (dummy,),
            output,
            opset_version=opset,
            do_constant_folding=True,
            input_names=list(inputs.keys()),
            output_names=list(outputs.keys()),
            dynamic_axes=dict(chain(inputs.items(), outputs.items())),
        )

        # Quantize model
        if quantize:
            if not ONNX_RUNTIME:
                raise ImportError(
                    'onnxruntime is not available "pip install onnxruntime"')

            output = self.quantization(output)

        if isinstance(output, BytesIO):
            # Reset stream and return bytes
            output.seek(0)
            output = output.read()

        return output



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['setfitonnx']

package_data = \
{'': ['*']}

install_requires = \
['joblib>=1.2.0,<2.0.0',
 'loguru>=0.6.0,<0.7.0',
 'onnx>=1.12.0,<2.0.0',
 'onnxruntime>=1.13.1,<2.0.0',
 'sclblonnx>=0.1.14,<0.2.0',
 'setfit>=0.4.1,<0.5.0',
 'skl2onnx>=1.13,<2.0']

setup_kwargs = {
    'name': 'setfitonnx',
    'version': '0.1.0',
    'description': 'Package to pack setfit model to onnx format',
    'long_description': '# SetFitONNX\n## Export the setfit model to ONNX format\n\n\n## Features\n\n- Mean Pooling Layer to ONNX Graph\n- Single ONNX Model combining both the sentence transformer & classification head\n\n## Installation\n```sh\npip install setfitonnx\n```\n## How to \n```python\nfrom setfitonnx import convert_onnx\n\n# Setfit Model directory\nmodel_path = "/home/mysetfit-model"\n# ONNX Output directory\noutput_dir = "/home/setfit-onnx-model"\n# Convert to ONNX\nconvert_onnx(model_path=model_path,output_dir=output_dir)\n\n```\n\n## License\n\nMIT',
    'author': 'Vimal Babu',
    'author_email': 'personal.vimal@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/vimal-b/setfitonnx',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
